from .entity import Ads, Ad
from .entity import Campaigns, Campaign
from .entity import Channels, ChannelItem
from .entity import Keywords, Keyword
from .entity import Placements, Placement
from .entity import Attribution
from .entity import Payload
from .common import Filter, Method, GroupBy, FilterCategory
from .common import AttributionModel,TypeReport
from .client import Client



import json
from datetime import datetime

from pydantic import BaseModel

from .common import AttributionModel, GroupBy, TypeReport, FilterCategory
from .utils import normalize_date


class ChannelItem(BaseModel):
    id: int
    title: str


class Channels(BaseModel):
    channels: list[ChannelItem]


class Channel(BaseModel):
    id: int
    title: str
    channel_id: int


class Placement(BaseModel):
    id: int
    title: str
    channel: Channel


class Placements(BaseModel):
    placements: list[Placement]


class Campaign(BaseModel):
    id: int
    placement_id: int
    title: str


class Campaigns(BaseModel):
    campaigns: list[Campaign]


class Keyword(BaseModel):
    id: int
    keyword: str


class Keywords(BaseModel):
    keywords: list[Keyword]


class Ad(BaseModel):
    id: int
    external_id: str
    placement_id: int
    campaign_id: int | None
    external_campaign_id: str | None
    type: str
    title: str | None
    text: str | None
    text1: str | None
    text2: str | None
    preview_url: str | None
    href: str | None
    device: str | None
    created_at: str


class Ads(BaseModel):
    ads: list[Ad]


class Attribution:

    def __init__(self, model_id: AttributionModel, period: int, with_direct: bool):
        self.model_id = model_id.value
        self.period = period
        self.with_direct = with_direct

    def __str__(self):
        return f"{self.model_id} {self.period} {self.with_direct}"


class Filter:
    def __init__(self, category: FilterCategory, value: int, operand: str = "="):
        self.category = category.value
        self.value = value
        self.operand = operand


class Payload:
    """Базовый класс параметров для генерации параметров к запросам"""

    def __init__(self, project: str, metrics: [str], datetime_from: datetime, datetime_to: datetime,
                 group_by: GroupBy, attribution: Attribution, type_report: TypeReport = TypeReport.AGGREGATED,
                 filters: list[Filter] | None = None, fields: list[str] | None = None):
        self.project = project
        self.metrics = ";".join(metrics) if metrics else ""
        self.datetime_from: str = normalize_date(datetime_from)
        self.datetime_to: str = normalize_date(datetime_to)
        self.group_by = group_by
        self.type = type_report
        self.filters = filters
        self.attribution = attribution
        self.fields = ";".join(fields) if fields else ""

    def __str__(self):
        return f"{self.datetime_from}//{self.datetime_to}"

    def to_json(self) -> str:
        payload = {"project": self.project,
                   "metrics": self.metrics,
                   "datetimeFrom": self.datetime_from,
                   "datetimeTo": self.datetime_to,
                   "groupBy": self.group_by.value,
                   "type": self.type.value}
        if self.filters:
            payload["filters"] = self.filters

        payload["attribution"] = self.attribution.__dict__

        if self.fields:
            payload["fields"] = self.fields
        return json.dumps(payload)



"""Словари используемые в модуле main."""
from collections import namedtuple
from enum import Enum


class Filter(Enum):
    SmartisID = 7071
    Channel = 1222
    Placement = 1223


MethodFields = namedtuple("MethodFields", "method location id")


class Method(Enum):
    get_projects = MethodFields("projects/get", "projects", "project")
    get_metrics = MethodFields("metrics/get", "metrics", "code")
    get_groupings = MethodFields("reports/getGroupings", "groupings", "code")
    get_attributions = MethodFields("reports/getModelAttributions", "modelAttributions", "id")
    get_channels = MethodFields("reports/getChannels", "", "")
    get_placements = MethodFields("reports/getPlacements", "", "")
    get_campaigns = MethodFields("reports/getCampaigns", "", "")
    get_ads = MethodFields("reports/getAds", "", "")
    get_keywords = MethodFields("reports/getKeywords", "", "")


class GroupBy(Enum):
    ADS = "ad_id"
    DAY = "day"
    PLACEMENT = "placement_id"
    CAMPAIGN = "campaigns"
    OBJECTS = "smartis_object"


class FilterCategory(Enum):
    SMARTIS_ID = 7071
    CHANNEL = 1222
    PLACEMENT = 1223


class AttributionModel(Enum):
    LAST_CLICK = 1  # Последнее касание
    FIRST_CLICK = 2  # Первое касание
    LINEAR = 3  # Линейное распределение
    BY_POSITION = 4  # На основе позиции
    FIRST_COMMUNICATION = 5  # Первое обращение
    LINEAR_BY_COMMUNICATION = 6  # Линейное распределение на обращениях
    LINEAR_WITH_POSTVIEW = 10  # Линейное распределение с учетом postview
    LAST_CLICK_WITH_POSTVIEW = 15  # Последнее касание с учетом postview
    FIRST_CLICK_WITH_POSTVIEW = 16  # Первое касание с учетом postview
    NOT_FIRST_NOT_LAST_CLICK = 17  # Не первое и не последнее
    LAST_COMMUNICATION = 22  # Последнее обращение
    BY_POSITION_WITH_POSTVIEW = 23  # На основе позиции с учетом postview



class TypeReport(Enum):
    RAW = "raw"
    AGGREGATED = "aggregated"



"""Дополнительные методы Smartis API"""
import json
import time
from enum import Enum

import requests
from pydantic import ValidationError, BaseModel

from .entity import Ads, Ad, Payload
from .entity import Campaigns, Campaign
from .entity import Channels, Placements
from .entity import Keywords, Keyword
from .common import Method
from .errors import ApiLimitError
import logging


class ContentType(Enum):
    application = "application/json"


class Client:
    TRY_REQUEST = 10
    REQUEST_PAUSE = 1
    FORCE_PAUSE = 600

    def __init__(self, api_key: str, dev: bool = False) -> None:
        self.api_key = api_key
        self.header = {'Authorization': f"Bearer {api_key}", "Content-Type": ContentType.application.value, }
        if dev:
            self.host = "https://dev.smartis.bi/api/"
        else:
            self.host = "https://my.smartis.bi/api/"

    def get_other(self, method: Method) -> []:
        response = self._prepare(method)
        return response.json()[method.value.location]

    def _prepare(self, method: Method, parameters: str | None = None) -> requests.Response:
        if parameters:
            resp = requests.post(f"{self.host}{method.value.method}", headers=self.header, data=parameters)
        else:
            resp = requests.post(f"{self.host}{method.value.method}", headers=self.header)

        if resp.status_code != 200:
            raise ConnectionError(f"status code: {resp.status_code}")

        return resp

    def get_channels(self) -> Channels:
        response = self._prepare(Method.get_channels)
        try:
            return Channels.model_validate(response.json())
        except ValidationError as e:
            raise ValueError from e

    def get_placements(self) -> Placements:
        response = self._prepare(Method.get_placements)
        try:
            return Placements.model_validate(response.json())
        except ValidationError as e:
            raise ValueError from e

    def get_campaigns(self, campaigns_ids: list) -> list[Campaign]:
        all_campaigns: list[Campaign] = []
        items: Campaigns = self._get_entity(campaigns_ids, Method.get_campaigns, Campaigns)
        all_campaigns.extend(items.campaigns)
        return all_campaigns

    def get_ads(self, ads_ids: list) -> list[Ad]:
        all_ads: list[Ad] = []
        items: Ads = self._get_entity(ads_ids, Method.get_ads, Ads)
        all_ads.extend(items.ads)
        return all_ads

    def _get_entity(self, ids: list, method: Method, model: BaseModel) -> BaseModel:
        ids = json.dumps({"ids": ids})
        response = self._prepare(method, parameters=ids)
        if response.status_code != 200:
            raise ConnectionError(f"status code: {str(response.status_code)}")
        try:
            return model.model_validate(response.json())
        except ValidationError as e:
            raise ValueError from e

    def get_keywords(self, keywords_ids: list) -> list[Keyword]:
        all_keywords: list[Keyword] = []
        items: Keywords = self._get_entity(keywords_ids, Method.get_keywords, Keywords)
        all_keywords.extend(items.keywords)
        return all_keywords

    def status_code_handler(self, response: requests.Response, retry: int) -> None:
        """Обработка статус кодов"""

        match response.status_code:
            case 200:
                return
            case 400 | 403:
                resp_body: dict = response.json()
                message = resp_body.get("error")
                raise ConnectionError(f"{response.status_code}:{response.text}. details: {message}")
            case 429:
                if int(response.headers['X-Ratelimit-Remaining']) == 0:
                    logging.warning(f"Status code: {response.status_code}. "
                                    f"Причина: {response.reason}. "
                                    f"Пауза: {response.headers['Retry-After']} с.")
                    raise ApiLimitError(int(response.headers['Retry-After']), response.text)
            case 500 | 502:
                logging.warning(f"Status code: {response.status_code}. "
                                f"Причина: {response.reason}. "
                                f"Пауза: 60 с.")
                time.sleep(60)
            case _:
                logging.warning(f" Необработанное исключение. "
                                f"Status code: {response.status_code}. "
                                f"Причина: {response.reason}. ")
                logging.info(response.headers)
                time.sleep(60)

    def get_report(self, payload: Payload, retry: int = TRY_REQUEST) -> requests.Response:
        """Выполнение запросов с обработкой ошибок"""
        payload_json = payload.to_json()
        try:
            time.sleep(self.REQUEST_PAUSE)
            answer = requests.post(f"{self.host}reports/getReport", headers=self.header, data=payload_json)
            self.status_code_handler(answer, self.TRY_REQUEST)
        except ApiLimitError as apiErr:
            if retry == 0:
                return self.retry_get_report(payload)
            logging.warning(f"error: {apiErr.message}")
            logging.info(f"Повтор запроса: {self.TRY_REQUEST - retry + 1}/{self.TRY_REQUEST}")
            return self.get_report(payload, retry=retry - 1)
        except Exception as er:
            if retry == 0:
                return self.retry_get_report(payload)
            logging.warning(f"Необработанное исключение. error: {er}")
            logging.info(f"Повтор запроса: {self.TRY_REQUEST - retry + 1}/{self.TRY_REQUEST}")
            return self.get_report(payload, retry=retry - 1)
        else:
            return answer

    def retry_get_report(self, payload: Payload) -> requests.Response:
        logging.info("Количество ошибок больше заданного! Пауза 10 мин")
        time.sleep(self.FORCE_PAUSE)
        return self.get_report(payload)



class ApiLimitError(Exception):
    def __init__(self, retry_after: int, *args):
        self.retry_after = retry_after
        self.message = args[0] if args else None

    def __str__(self):
        return f'ApiLimitError, {self.message} ' if self.message else 'ApiLimitError'
